# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 17/12/2016
@version: 0.1

Defines custom exceptions used by Pwf or that can be used
in the view fucntion.
"""


class HTTPException(Exception):

    code = None
    description = None

    def __init__(self, code=None, description=None):
        Exception.__init__(self)

    @property
    def respcode(self):
        return self.code


class NotFound(HTTPException):
    code = 404
    description = ''


class MethodNotAllowed(HTTPException):
    code = 405
    description = ''


class InternalError(HTTPException):
    code = 500
    description = ''


class RequestEntityTooLarge(HTTPException):
    code = 413
    description = ''




# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 18/12/2016
@version: 0.1
"""

from time import gmtime
from collections import namedtuple
from datetime import datetime, timedelta

def create_cookie(key, value='', path='/', expires=None, max_age=None,
            domain=None, secure=False, httponly=False):
    """Generates a WSGI compatible cookie dictionary. Key and value are 
    requred and we then loop through the additional ones to check what
    to add and how.
    """
    if isinstance(max_age, timedelta):
        max_age = (max_age.days * 60 * 60 * 24) + max_age.seconds
  
    if expires and not isinstance(expires, str):
        expires = format_date(expires)

    attributes = (('Domain', domain, True),
                  ('Expires', expires, False,),
                  ('Max-Age', max_age, False),
                  ('Secure', secure, None),
                  ('HttpOnly', httponly, None),
                  ('Path', path, False))
       
    buf = [key + '=' + value]
    for name, value, use_value in attributes:
        
        # If use_value is None but value is set we only add the
        # name to the cookie (Example: Secure rather than Secure=True)
        if use_value is None:
            if value:
                buf.append(name)
            continue # pragma: no cover

        if value is None:
            continue
        
        if not isinstance(value, (bytes, bytearray)):
            value = bytes(value)

        attribute_string = '{}={}'.format(name, value)
        buf.append(bytes(attribute_string))
    
    cookie_string = '; '.join(buf)
    cookieheader = {'Set-Cookie': cookie_string}

    return cookieheader



def format_date(date):
    """Format to Wed, 21-Dec-2016 21:28:34 GMT.
    Accepts int, float or datetime object.
    """
    if date is None:
        date = gmtime()
    elif isinstance(date, datetime):
        date = date.utctimetuple()
    elif isinstance(date, (int, float)):
        date = gmtime(date)

    return '%s, %02d%s%s%s%s %02d:%02d:%02d GMT' % (
        ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')[date.tm_wday],
        date.tm_mday, '-',
        ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
        'Oct', 'Nov', 'Dec')[date.tm_mon - 1],
        '-', str(date.tm_year), date.tm_hour, date.tm_min, date.tm_sec
        )







# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 17/12/2016
@version: 0.1
"""


import shutil
import json
from tempfile import TemporaryFile
from io import BytesIO


class FileWrapper(object):
    """Wrapper around a wsgi.input filestream. 
    It adds functionality for retrieving the filename
    and saving the file to disk.
    
    This can also be used to wrap binary data and treat
    it like a file object.
    """

    def __init__(self, file, filename=None, name=None,
                 mimetype=None, headers=None):

        if isinstance(file, str):
            self.file = BytesIO(file)
        else:
            self.file = file

        self.filename = filename
        self.name = name
        self.headers = headers
        self.mimetype = mimetype

        self.buffer_size = 16384

    def save(self, dest):
        """Copies the content of the filestream into a file
        in the specified destination. We copy the file
        over in chunks equal to the buffer_size.

        Args:
            dest (str): The path where we save the file.

        Example usage:
            my_file = request.files['deathstarplans']
            my_file.save('/path/to/save/secret-plans.jpg')

        """
        if isinstance(dest, (str, unicode)):
            dest = open(dest, 'wb')
        try:
            shutil.copyfileobj(self.file, dest, self.buffer_size)
        finally:
            dest.close()

    def seek(self, pos):
        """Seek the file to the specified position (pos)"""
        self.file.seek(pos)

    def read(self, length=-1):
        """Read the file to the specified length or EOF"""
        return self.file.read(length)
        
    def close(self):
        """Close the file. If you try to access the FileWrapper
        object after it's been closed it will raise an ValueError.
        """
        try:
            self.file.close()
        except Exception:
            pass

    def readline(self):
        return self.file.readline()


class Config(dict):
    """Subclass of dictionary adding the capability of loading
    the configuration from a json file. The config objects
    can be accessed through "app.config".

    Example:

        app.config['DB_USERNAME'] = 'mos_eisley'
        app.config['DB_PASSWORD'] = 'wretchedhiveofscumandvillany'

    """

    def __init__(self):
        dict.__init__(self or {})

    def from_json_file(self, path):
        with open(path) as json_file:
            data = json.loads(json_file.read())
       
        for k, v in data.iteritems():
            self[k] = v




# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 07/01/2017
@version: 0.1
"""


class AppStack(object):
    
    def __init__(self):
        self.items = []

    def push(self, item):
        self.items.append(item)

    def pop(self):
        return self.items.pop()

    @property 
    def top(self):
        try:
            return self.items[-1]
        except (AttributeError, IndexError):
            return None

    @property    
    def is_empty(self):
        return (self.items == [])

    def reset(self):
        self.items = []
    

_app_stack = AppStack()



# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 17/12/2016
@version: 0.1
"""

import re
from wrappers import Config
from request import Request
from response import Response
from functools import wraps
from stack import _app_stack
from exceptions import HTTPException, MethodNotAllowed, NotFound, InternalError
from utils import log
import wsgiref.simple_server


class Pwf(object):
    """This is the central object you instantiate and interact with in your
    program. Once created it can be used in the view to access things
    like cookies, headers, response codes etc.

    Example usage to define a view:
    
        app = Pwf()

        app.routes('/')
        def index(request):
                return 'Hello world'


    The above example instantiates the Pwf object, defines a route to '/'
    through the app.routes decorator and returns 'Hello World' from the
    view function.
    """
    
    def __init__(self):
        self.routes = []
        self.config = Config()
        self.first_funcs = {}
        self.last_funcs = {}
        self.errorhandlers = {}
        _app_stack.push(self)
    
    def run(self, host='127.0.0.1', port=5000):
        """Serves the PWF app using the wsgiref simple server. Use this
        only for quick testing in development.
        """
        httpd = wsgiref.simple_server.make_server('', port, self)
        log("PWF now running on http://%s:%s/" % (host, port,))
        httpd.serve_forever()

    def build_route_pattern(self, url):
        """Regex to find path variables in path"""
        route_regex = re.sub(r'(<\w+>)', r'(?P\1.+)', url)
        return re.compile("^{}$".format(route_regex))

    def get_route_match(self, path):
        """Match a path to an entry in self.routes and return variables,
        supported methods and view function
        """
        for route_pattern, methods, group, view_function in self.routes:
            m = route_pattern.match(path)
            if m:
                return m.groupdict(), methods, group, view_function
            
        return None

    def route(self, url, methods=['GET'], group=None):
        """This function is used as a decorator for each view function
        to define the route associated with that view.
        
        Example:
            @app.route('/profile')
            def profile(request):
                return 'This is the profile page'

        It adds the path along with the view function itself, the request
        method and any route variables to self.routes for the path_dispatcher
        to user.
        """
        def decorate(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)
        
            route_pattern = self.build_route_pattern(url) 
            self.routes.append((route_pattern, methods, group, f))
            return wrapper

        return decorate

    def path_dispatch(self, request, make_response):
        """Dispatches the request, does pre and post processing and
        dispatches any exceptions to the exception handler.
        """
        try:
            rv = self.preprocess_request(request)
            if rv is None:
                rv = self.process_request(request)
        except Exception as e:
            rv = self.handle_exception(e, make_response)

        return self.postprocess_request(rv, make_response)

    def dispatch_request(self, environ, make_response):
        """Instantiate a new Request object based on environ,
        pass it to the path dispatcher to get the response instance.
        """
        request = Request(environ)
        response = self.path_dispatch(request, make_response)
        return response

    def make_response(self, data=None):
        """Called from the view function to create and return
        a response object. This can then be used to add cookies,
        headers and more.
        """
        response = Response(data=data)
        return response

    def first(self, group=None):
        """Decorator used to specify functions to be executed
        before a request. If these functions return a value
        that will be used in the response rather than that of
        the view.
        
        Example usage to return "405 Method Not Allowed" for all
        request with content type text/plain:
        
        @app.first()
        def check_content_type(request):
            if request.headers['CONTENT_TYPE'] == 'text/plain':
                resp = app.make_response()
                resp.code = 405
                return resp
        """
        def wrapper(f):
            self.first_funcs.setdefault(group, []).append(f)
            return f
        return wrapper

    def last(self, group=None):
        """Decorator used to specify functions to be executed
        after each request. If these functions return a value
        that will be used in the response.
        
        Example usage to add a header to all responses:

        @app.last()
        def add_header(response):
            response.headers['Content-Type'] = 'application/json'
            return response
        """
        def wrapper(f):
            self.last_funcs.setdefault(group, []).append(f)
            return f
        return wrapper

    def error(self, code):
        """
        Define an error handler function to return custom data
        for different error codes.
        
        Example:
            
            @app.error(405)
            def handle_forbidden():
                return "I'm sorry Dave, I'm afraid i can't do that"
        """
        def wrapper(f):
            self.errorhandlers[code] = f
        return wrapper

    def preprocess_request(self, request):
        """The first step of the request to response cycle.
        Here we execute any general @app.first functions and
        if a return value is given return it to the server.
        """
        #: Execute any general "first" functions
        rv = self.__execute_first(request)
        if rv is not None:
            return rv

    def process_request(self, request):
        """Gets the requested path from the headers and
        matches it to a route, function, group and method stored in
        self.routes. The view funtion itself then gets executed to generate
        the response-data.

        If the requested path is not found in self.routes or the
        request method used is not supported by the view we return
        an error Response object.

        We also execute any "first" and "last" functions.
        """
        path = request.headers['PATH_INFO']
        method = request.headers['REQUEST_METHOD']

        #: Match to a route
        route_match = self.get_route_match(path)

        if route_match:
            kwargs, methods, group, view_function = route_match

            #: Execute any first functions for route group
            first_group_rv = self.__execute_first_groups(request, group)
            if first_group_rv is not None:
                return first_group_rv
            
            if method not in methods:
                raise MethodNotAllowed()
            else:
                rv = view_function(request, **kwargs)

            #: Execute any last functions for a route group. If we don't want
            #: functions without a group to be executed on top of the ones with
            #: a group, we should build and return the response here.
            last_group_rv = self.__execute_last_groups(rv, group)      
            if last_group_rv is not None:
                return last_group_rv
            
        else:
            raise NotFound()
            
        return rv

    def postprocess_request(self, rv, make_response):
        """The last step of the request to response cycle.
        Here we execute any general @app.last functions and
        return a valid response object.
        """
        response = self.__build_response(rv, make_response)
        
        try:
            last_rv = self.__execute_last(response)
        except Exception as e:
            last_rv = self.handle_exception(e, make_response)

        if last_rv is not None:
            return last_rv

        return response
    
    def handle_exception(self, e, make_response):
        """Builds the appropriate response object
        based on what exception was raised.
        
        If debug is true we don't handle the exception
        but just re-raise it again.
        """
        if self.config.get('DEBUG', None):
            raise e

        if isinstance(e, (NotFound, MethodNotAllowed)):
            rv = self.__error_return(make_response, e.respcode)
        elif isinstance(e, Exception):
            rv = self.__error_return(make_response, 500)
        
        return rv

    def __build_response(self, data, make_response, code=200):
        """Generaes the final response object.

        If the view function returns a Response object (The Response
        object was created in the view) we attach the make_response
        to it. If the view function return is standard data (and headers)
        we create the Response object here.
        """
        if isinstance(data, Response):
            response = data
            response.make_response = make_response # Attach the WSGI make_response object
        else:
            response = Response(make_response, code, data=data)
        
        return response

    def __execute_first_groups(self, request, group):
        """Execute only the first functions attached to a group."""
        if group is not None:
            rv = self.__execute_first(request, group)
            return rv

    def __execute_last_groups(self, response, group):
        """Execute only the last functions attached to a group-"""
        if group is not None:
            rv = self.__execute_last(response, group)
            return rv

    def __execute_first(self, request, group=None):
        """Runs through and executes any functions
        in the self.first_funcs dictionary.
        """
        funcs = self.first_funcs.get(group, ())
        for func in funcs:
            rv = func(request)
            if rv is not None: 
                return rv

    def __execute_last(self, response, group=None):
        """Runs throug and executes any functions
        in the self.last_funcs dictionary.
        """
        if not isinstance(response, Response):
            response = Response(data=response)

        funcs = self.last_funcs.get(group, ())
        for func in funcs:
            rv = func(response)
            if rv is not None:
                return rv

    def __error_return(self, make_response, code=404):
        """Checks if an error handler was defined by the
        user and in that case returns that as the response.
        If no handler was defined we return an empty response.
        """
        errorhandler = self.errorhandlers.get(code, None)

        if callable(errorhandler):
            rv = errorhandler()
        else:
            rv = None
        
        if rv is not None:
            response = self.__build_response(rv, make_response, code=code)
        else:
            response = Response(make_response, code)
        return response
        
    def __call__(self, environ, make_response):
        """Gets executed every time an instance of the class
        gets called. 
        
        Calls dispatch_request with the standard WSGI objects
        environ and make_request to get back a response.

        It then calls render() on the response to render it 
        back to the server.
        """
        resp = self.dispatch_request(environ, make_response)
        return resp.render()

    def __repr__(self):
        return '%s()' % self.__class__.__name__



# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 09/01/2017
@version: 0.1

Implements a set of utilities used throughout Pwf but some
might be useful to use in other circumstances as well.
"""
import logging


class cached_property(property):
    """Extends the built in class property with cache
    functionality. It is used in the same way you would
    the regular property decorator.

    The return value is calculated once and then stored
    in the object. Any subsequent access will return 
    the cached value.

    Example:

    class MyClass(object):
        
        @cached_property
        def method(self):
            return 2

    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = func.__name__
        #self.__module__ = func.__module__
        #self.__doc__ = func.__doc__
        self.func = func

    def __set__(self, obj, value):
        print obj
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None: # pragma: no cover
            return self

        value = obj.__dict__.get(self.__name__, None)
        if value is None:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value



def log(info):
    """TODO: Implement a real loggin solution"""
    print info






# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 18/12/2016
@version: 0.1

Features implemented here are series of utilities not used by
PWF directly but that might come in handy. They are not related
to the PWF core or specific to WSGI.
"""

import json
import time
from functools import wraps


def render_json(data, headers=None):
    """Generates a valid JSON response
    along with the correct content-type header.

    @param data: A dict that will be converted to JSON.
    @param headers: A dict of additional reponse headers.

    Note: If the header param already contains a content-type
    it will be removed and replaced by the 'application/json' header.
    """
    if isinstance(headers, dict):
        content_types = [key for key in headers.keys() if key.lower() == 'content-type']
        for key in content_types:
            headers.pop(key)
    else:
        headers = {}

    # Turn header values into strings
    for header, value in headers.iteritems():
        if not isinstance(value, str):
            headers[header] = bytes(value)

    headers['content-type'] = 'application/json'
    return json.dumps(data), headers


def timed(f):
    """A simple decorator that calculates the amount of time
    a view function takes to execute (in milliseconds). The
    result is printed to sys.stdout.
    """
    @wraps(f)
    def wrapper(*args, **kwds):
        start = time.time()
        result = f(*args, **kwds)
        elapsed = float((time.time() - start) * 1000)
        print '{0} took {1:.2f} ms to finish'.format(f.__name__, elapsed)
        return result
    return wrapper





# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 17/12/2016
@version: 0.1
"""

import cgi
import urlparse
import json
import Cookie
from io import BytesIO
from tempfile import TemporaryFile

from wrappers import FileWrapper
from utils import cached_property
from stack import _app_stack
from exceptions import RequestEntityTooLarge


class Request(object):
    """Create a Request object populated with the contents of
    the WSGI environ dictionary.

    This object gets passed to the user defined view function
    to be consumed. 
    """

    def __init__(self, environ):
        """Set object variables that will be accessible
        in the view function.

        self.stream holds the cached wsgi.input as a file-like object.
        """
        self.stream = self.__cache_stream(environ)
        self.environ = environ
        self.headers = self.__parse_headers(environ)
        self.content_length = self.__parse_content_length(environ)
        self.mimetype, self.options = self.__parse_content_type(environ)
        self.method = self.__parse_method(environ)
        self.cookies = self.__parse_cookies(environ)
        self.query = self.__parse_query(environ)
        #self.json = None
        
        # Defines methods used by __parse_data depending on mime-type
        self.parse_methods = {
            'multipart/form-data': self.__parse_form,
            'application/x-www-form-urlencoded': self.__parse_form
            }

    def __parse_cookies(self, environ):
        """Get cookies from environ and return a dict with 
        keys and values.
        """
        parsed_cookies = {}
        cookies = Cookie.SimpleCookie(environ.get('HTTP_COOKIE', ''))
        for key in cookies:
            parsed_cookies[key] = cookies[key].value

        return parsed_cookies

    def __parse_content_type(self, environ):
        """Parses the HTTP Content-Type header into a typle of
        mimetype and a dictionary of any options.
        
        For example the content-type 'multipart/form-data; charset=UTF-8'
        is parsed into ('multipart/form-data', {'charset': 'utf-8'})
        
        The word "Content-Type" is used to refer to the HTTP header
        with both MIME and options, while "mimetype" refers to the
        MIME value only.
        """
        ct_value = environ.get('CONTENT_TYPE', '').lower()
        if not ct_value:
            return '', {}

        mimetype, options = cgi.parse_header(ct_value)
        return mimetype, options
        

    def __parse_headers(self, environ):
        """We parse the headers from the WSGI environ and
        return only the wanted ones defined in wanted_headers list
        plus any HTTP headers.
        """
        length = environ.get('CONTENT_LENGTH', 0)
        headers = { 'CONTENT_LENGTH': 0 if not length else int(length) }

        wanted_headers = ['REQUEST_METHOD', 'PATH_INFO', 'REMOTE_ADDR', 
                'REMOTE_HOST', 'CONTENT_TYPE']

        for k, v in environ.items():
            if k in wanted_headers or k.startswith('HTTP'):
                headers[k] = v
        
        # TODO: Add support for parsed headers. Parse
        # "HTTP_X_CUSTOM-HEADER" to "X-Custom-Header"

        return headers

    def __parse_content_length(self, environ):
        try:
            cl = int(environ.get('CONTENT_LENGTH', 0))
        except ValueError:
            cl = 0

        return cl
    

    def __parse_method(self, environ):
        """Return the current request method as a string.
        For example 'GET' or 'OPTIONS'
        """
        request_method = environ['REQUEST_METHOD']
        return request_method

    def __parse_query(self, environ):
        """Parse a query string into a dictionary."""
        qs = environ.get('QUERY_STRING', None)
        query = urlparse.parse_qs(qs)

        # Only select the first value for any give query variable.
        # Perhaps support multiple values in the future 
        unique_variables = {k: v[0] for k, v in query.items()}
        return unique_variables
    
    def __parse_data(self, environ, parse_form=False):
        """Parses the data from the WSGI environ object depending on
        the mimetype. If no parse method is defined for a
        particular mimetype we fall back to return the
        data as a string.
        """

        length = self.content_length

        if parse_form is not True:
            return self.get_stream.read(length)

        parse_method = self.parse_methods.get(self.mimetype, None)

        if parse_method is not None:
            data = parse_method(environ)
        else:
            data = self.get_stream.read(length)

        return data


    def __parse_form(self, environ):
        """Parse form data. If a file is included we wrap it using
        the FileWrapper and add it to self.files. If it's regular
        form data we add the key and value to the data dict.
        """
        form = {}
        files = {}
        req_d = self.__dict__
        req_d['form'], req_d['files'] = form, files

        env_data = cgi.FieldStorage(self.get_stream, environ=environ,
                keep_blank_values=True)
        
        if not env_data:
            return form

        for k in env_data.list:
            # NOTE: Perhaps add support application/x-www-form-urlencoded
            # in the future.
            if isinstance(k, cgi.MiniFieldStorage):
                return form

            if k.filename:
                headers = dict(k.headers)
                filewrapper = FileWrapper(k.file, k.filename, 
                        k.name, mimetype=k.type, headers=headers)

                files[k.name] = filewrapper
            else:
                form[k.name] = k.value

        return form

    @cached_property
    def data(self):
        """Returns any request data as a form object or
        string, depending on the mimetype"""
        data = self.__parse_data(self.environ, parse_form=True)
        return data

    @cached_property
    def files(self):
        """Returns any uploaded files. Empty if no files where pared"""
        self.__parse_form(self.environ)
        return self.files

    @cached_property
    def form(self):
        """Returns any form data in the request. Emtpy if no form as parsed"""
        self.__parse_form(self.environ)
        return self.form

    @property
    def get_stream(self):
        """Returns the cached stream as a file-like object"""
        self.stream.seek(0)
        return self.stream

    @cached_property
    def json(self):
        """If the mimetype is application/json, parse self.data and
        return a python dictionary"""

        if not self.data:
            return None
    
        if not 'application/json' in self.mimetype:
            return None
       
        try:
            data = json.loads(self.data)
        except ValueError:
            return None

        return data
    
  
    def __cache_stream(self, environ):
        """Caches the query stream so it can be accessed
        multiple times. If the stream was is cached we read
        envrion['wsgi.input'] and store it.
        
        For large files the stream is stored in a temporary file.
        For smaller files (< 500KB) we store the data in memory as a
        BytesIO object.
        """
        try:
            content_length = int(environ.get('CONTENT_LENGTH', 0))
        except ValueError:
            content_length = 0

        stream = environ['wsgi.input'].read(content_length)
        
        # If the stream is large we use a temporary file. If not we
        # just load it into memory
        if content_length > 1024 * 500:
            _stream_cache = TemporaryFile('wb+')
            _stream_cache.write(stream)
            _stream_cache.seek(0)
        else:
            _stream_cache = BytesIO(stream)

        return _stream_cache


    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.__dict__)



# -*- coding: utf-8 -*-
"""
@author: Victor Kohler
@since: date 17/12/2016
@version: 0.1
"""

import httplib
import helpers
import cookies

class Response(object):
    """Used to set and return data back to the server. The Response object
    can be instantiated in one of two ways:

    1. It is created manually in the view function to add custom headers, 
    cookies or response codes.
    
    2. It gets created automatically once the view function returns
    some data.

    Create the response object from the view using app.make_response(data)
    """

    def __init__(self, make_response=None, code=200, data=''):
        """For the view data we're currently supporting either a tuple with
        both the returned data and a dictionary of headers or just the 
        returned data.
        """
        if isinstance(data, tuple):
            self.data = data[0]
            headers = data[1]
        else:
            self.data = data
            headers = {}

        self.headers = headers
        self.code = code
        self.make_response = make_response

    def set_cookie(self, key, value='', path='/', expires=None, max_age=None,
                domain=None, secure=False, httponly=False):
        """Creates a cookie dictionary and adds it to the headers.
        This function is ment to be used in the view function:
        
            resp = make_response(data)
            resp.set_cookie('key', 'value')
        """
        cookie = cookies.create_cookie(key, value, path, expires, max_age,
                domain, secure, httponly)

        #TODO: Handle multiple cookies
        self.headers.update(cookie)

    def render(self):
        """Renders the final response back to the server with status code,
        headers and data. It aslo transform headers and codes into
        a WSGI compatible format.
        
        If status code is 5xx or 4xx no view data is returned.
        """

        # If the content type is not specified, we set
        # it to text/html as the default
        if 'content-type' not in map(lambda x:x.lower(), self.headers):
            self.headers['Content-Type'] = 'text/html'

        # Set headers as list of tuples
        self.headers = [(k, v) for k, v in self.headers.items()]

        # httplib.responses maps the HTTP 1.1 status codes to W3C names.
        # Output example: '200 OK' or '404 Not Found'
        resp_code = '{} {}'.format(self.code, httplib.responses[self.code])

        if str(self.code)[0] in ['4', '5'] and not self.data:
            self.make_response(resp_code, self.headers)
            return resp_code.encode('utf-8')

        try:
            data = bytes(self.data).encode('utf-8')
        except UnicodeDecodeError:
            data = bytes(self.data)
        
        self.make_response(resp_code, self.headers)
        return data

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.__dict__)



from setuptools import setup, find_packages

setup(
    name='pwf',
    version='0.1.0',
    description='A Python Web Framework',
    long_description='',
    url='https://github.com/victorkohler/pwf',
    author='Victor Kohler',
    author_email='victor@houseofradon.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='pwf framework web development',
    packages=['pwf'],
    install_requires=[],
    entry_points={
        'console_scripts': [
        'pwf=pwf:main',
        ],
    },
)
